#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <pthread.h>
#include <signal.h>
#include <semaphore.h>
#include <fcntl.h>
#include <sys/stat.h>

#include "daware.h"

extern const char*
get_container_name();

#define PROC_DAWARE_PATH "/proc/daware"

#define DEBUG_LOG(fmt, ...)
//#define DEBUG_LOG(fmt, ...) printf(fmt, ##__VA_ARGS__)

#define PTHREAD_IS_ALIVE(t) (pthread_kill(t, 0) == 0 ? 1 : 0)

typedef struct _DTerminmation {
    int signo;
    int (*func)(int, void*);
    void *arg;
} _DTermination_t;

static _DTermination_t **cblist = NULL;

static void
cblist_add(_DTermination_t *d)
{
    if (cblist == NULL) {
        cblist = (_DTermination_t **)malloc(sizeof(_DTermination_t) * SIGRTMAX);
        if (cblist == NULL) {
            //daware_termination_unregister_callback((DTermination *)d);
            //return NULL;
            return;
        }
        memset(cblist, 0, sizeof(_DTermination_t) * SIGRTMAX);
    }
    if (0 < d->signo && d->signo <= SIGRTMAX) {
        cblist[d->signo - 1] = d;
    }
}

static void
cblist_remove(_DTermination_t *d)
{
    if (!cblist || !d) return;
    if (0 < d->signo && d->signo <= SIGRTMAX) {
        if (d == cblist[d->signo - 1]) {
            cblist[d->signo - 1] = NULL;
        }
    }
}

static void
cblist_exec_callback(int signo)
{
    if (0 < signo && signo <= SIGRTMAX) {
        _DTermination_t *d = cblist[signo - 1];
        if (d) {
            d->func(signo, d->arg);
        }
    }
}

static int
cblist_is_exist(_DTermination_t *d)
{
    if (!cblist || !d) return 0;
    if (d->signo <= 0 && SIGRTMAX < d->signo) return 0;
    return (d == cblist[d->signo - 1]);
}

static int
cblist_count()
{
    if (!cblist) return 0;

    int i;
    int count = 0;
    for(i = 0; i < SIGRTMAX; i++) {
        if (cblist[i]) count++;
    }
    return count;
}

int *pipefd = NULL;

static int
pipeInit()
{
    int ret = -1;
    pipefd = (int *)malloc(sizeof(int) * 2);
    if (pipefd) {
        if ((ret = pipe(pipefd)) == -1) {
            perror("pipe");
        }
    }
    return ret;
}

static int
pipeRecv(int *signo)
{
    int ret = -1;
    if (pipefd && pipefd[0]) {
        if ((ret = read(pipefd[0], signo, sizeof(int))) == -1) {
            perror("read");
        }
    }
    return ret;
}

static int
pipeSend(int *signo)
{
    int ret = -1;
    if (pipefd && pipefd[1]) {
        if ((ret = write(pipefd[1], signo, sizeof(int))) == -1) {
            perror("write");
        }
    }
    return ret;
}

static int
pipeFinish()
{
    if (pipefd) {
        int i;
        for (i = 0; i < 2; i++) {
            if (pipefd[i]) {
                if (close(pipefd[i]) == -1) {
                    perror("close");
                }
            }
        }
        free(pipefd);
        pipefd = NULL;
    }
    return 0;
}

sem_t *sem = NULL;
int signo_for_sem = 0;

static int
semInit()
{
    int ret = -1;
    sem = (sem_t *)malloc(sizeof(sem_t));
    if (sem) {
        if ((ret = sem_init(sem, 0, 0)) == -1) {
            perror("sem_init");
        }
    }
    return ret;
}

static int
semRecv(int *signo)
{
    int ret = -1;
    if (sem) {
        if ((ret = sem_wait(sem)) == -1) {
            perror("sem_wait");
        }
        *signo = signo_for_sem;
    }
    return ret;
}

static int
semSend(int *signo)
{
    int ret = -1;
    if (sem) {
        signo_for_sem = *signo;
        if ((ret = sem_post(sem)) == -1) {
            perror("sem_post");
        }
    }
    return ret;
}

static int
semFinish()
{
    int ret = -1;
    if (sem) {
        if ((ret = sem_destroy(sem)) == -1) {
            perror("sem_destroy");
        }
        free(sem);
        sem = NULL;
    }
    return ret;
}

struct {
    int (*init)();
    int (*recv)(int*);
    int (*send)(int*);
    int (*finish)();
    char *type;
} commufunc[] = {
    { pipeInit, pipeRecv, pipeSend, pipeFinish, "pipe" },
    { semInit,  semRecv,  semSend,  semFinish,  "sem" }
};

#define FUNC_TYPE 1


struct timespec tp_start;
struct timespec tp_end;
clockid_t clockid = CLOCK_PROCESS_CPUTIME_ID;

static void
debug_start_timer()
{
    clock_gettime(clockid, &tp_start);
}

static void
debug_stop_timer()
{
    clock_gettime(clockid, &tp_end);
    double time = tp_end.tv_sec - tp_start.tv_sec
        + (double)(tp_end.tv_nsec - tp_start.tv_nsec) / 1000000000;
    DEBUG_LOG("(%s)time: %.9f\n", commufunc[FUNC_TYPE].type, time);
}


pthread_t *tid = NULL;

static void *
thread_func(void *arg)
{
    int signo;
    while(1) {
        if (commufunc[FUNC_TYPE].recv(&signo) == -1) {
            return NULL;
        }
        debug_stop_timer();
        cblist_exec_callback(signo);
    }
    return NULL;
}

static int
stop_thread()
{
    int ret = -1;
    if (tid) {
        if (PTHREAD_IS_ALIVE(*tid)) {
            if((ret = pthread_cancel(*tid)) != 0) {
                fprintf(stderr, "pthread_cancel: error %d\n", ret);
            }

            if((ret = pthread_join(*tid, NULL)) != 0) {
                fprintf(stderr, "pthread_join: error %d\n", ret);
            }
        }
    }

    commufunc[FUNC_TYPE].finish();

    return 0;
}

static int
start_thread()
{
    int ret = -1;

    stop_thread();

    if (commufunc[FUNC_TYPE].init() == -1) {
        return -1;
    }

    tid = (pthread_t *)malloc(sizeof(pthread_t));
    if (tid == NULL) {
        return -1;
    }

    if ((ret = pthread_create(tid, NULL, thread_func, NULL)) != 0) {
        fprintf(stderr, "pthread_create: error %d\n", ret);
        return -1;
    }

    return 0;
}

static int
proc_daware_write(const char *msg, int size)
{
    int fd = open(PROC_DAWARE_PATH, O_WRONLY);
    if (fd == -1) {
//        perror("open");
        return -1;
    }

    if (write(fd, msg, size) == -1) {
        perror("write");
        return -1;
    }

    if (close(fd) == -1) {
        perror("close");
        return -1;
    }

    return 0;
}

static int
register_pid()
{
    char msg[32];
    int size = snprintf(msg, sizeof(msg), "%s %d",
             get_container_name(), getpid());

    return proc_daware_write(msg, size);
}

static int
unregister_pid()
{
    char msg[32];
    int size = snprintf(msg, sizeof(msg), "-d %s %d",
             get_container_name(), getpid());

    return proc_daware_write(msg, size);
}

static void
signal_handler(int signo) {
    debug_start_timer();
    commufunc[FUNC_TYPE].send(&signo);
}

void
daware_termination_unregister_callback(DTermination *dterm)
{
    if (dterm) {
        _DTermination_t *d = (_DTermination_t *)dterm;

        if (cblist_is_exist(d)) {
            struct sigaction sa;
            sa.sa_handler = SIG_DFL;
            sigemptyset(&sa.sa_mask);
            sigaction(d->signo, &sa, 0);

            cblist_remove(d);
            DLOG_INFO("event=DAWARE Unregister");
        }
        free(dterm);
    }

    stop_thread();

    if (!cblist_count()) {
        unregister_pid();
    }
}

DTermination *
daware_termination_register_callback(int signo, int (*func)(int, void*), void *arg)
{
    if (!cblist_count()) {
        register_pid();
    }

    if (signo <= 0 && SIGRTMAX < signo) return NULL;

    _DTermination_t *dterm = (_DTermination_t *)malloc(sizeof(_DTermination_t));
    if (dterm == NULL) {
        return NULL;
    }
    dterm->signo = signo;
    dterm->func  = func;
    dterm->arg   = arg;

    if (start_thread() != 0) {
        daware_termination_unregister_callback((DTermination *)dterm);
        return NULL;
    }

    cblist_add(dterm);
    DLOG_INFO("event=DAWARE Register");

    struct sigaction sa;
    sa.sa_handler = signal_handler;
    sa.sa_flags   = SA_RESTART;
    sigemptyset(&sa.sa_mask);
    sigaction(signo, &sa, 0);

    return (DTermination *)dterm;
}
